# -*- cperl -*-

package GDT_M_Edit;
BEGIN { import TredMacro; }
use strict;
use warnings;
use DBI;
use Data::Dumper;
use Lemmatizer;
use Tk::Table;
use TrEd::Basics;
#binding-context GDT_M_Edit

=head1 GDT_M_Edit

=head2 DESCRIPTION

Provides commands for editing tags in GDT.

Macro edit_tag, bound to key [t] opens tag editor.

The tag editor consists of a label displaing a partial tag and two
list-boxes. The top of the window shows current partial tag, the left
list-box offers immediate extensions of the partial tag and the right list-box
offers all possible completions of the partial tag.

The list-boxes can be navigated by mouse, or by keyboard.
Double-clicks, [Return] and [Space] confirm the selected item. If it
completes the tag, the tag is assigned. Otherwise the value is
appended to the partial tag and the list-box are updated accordingly.
If all immediate extensions of the tag complete the tag, the
second-listbox is automatically focused.

Pressing [Backspace], the arrow button, or choosing '..' from the left
list-box removes the trailing portion of the partial tag.

Items can also be selected by typing. If there is only one item that mathes
with the typed letters, the item is automatically confirmed.

=cut

# detect GDT files and apply this context
push @TredMacro::AUTO_CONTEXT_GUESSING, sub {
  my ($hook)=@_;
  if($grp->{FSFile}->FS->exists('afun')
     and
     grep { $_ eq 'IObj' } $grp->{FSFile}->FS->listValues('afun') ) {
    return 'GDT_M_Edit';
  }
  return;
};

my %unknownTokens;



# apply GDT stylesheet
sub switch_context_hook {
  SetCurrentStylesheet('GDT_M_Edit');
  # SetCurrentStylesheet('GDT_POS_Vs_Deps_Check');
  Redraw() if GUI();

  # Modify default TrEd's keybinding.
  # Up/Down move to previous/next node
  my @keys = qw(Up Down);
  my %normal = map { $_ => OverrideBuiltinBinding('*',$_) } @keys;
  my %shift = map { $_ => OverrideBuiltinBinding('*',"Shift+$_") } @keys;
  for (@keys) {
    OverrideBuiltinBinding(__PACKAGE__, $_, $shift{$_});
    OverrideBuiltinBinding(__PACKAGE__, "Shift+$_", $normal{$_});
  }
}

{
  # return a tagset as an array reference
  # (loading from source file just once)
  # This is loaded when TrEd starts
  my $tagset;
  sub get_tagset {
    unless ($tagset) {
      my $tagset_file = FindInResources('gdt/tagset.txt');
      open my $fh, '<', $tagset_file || die "Failed to open tagset file $tagset_file: $!";
      $tagset=[];
      while (<$fh>) {
	chomp;
	push @$tagset,$_ if /\S/;
      }
      close($fh);
    }
    return $tagset;
  }
}

{
  # return the DB configuration settings as a hash reference
  # (loading from source file just once)
  # This is loaded when TrEd starts
  my $dbConfig;
  my $dbConfigTemp;
  sub getDbConfig {
    unless ($dbConfig) {
      my $dbConfig_file = FindInResources('gdt/dbConfig.txt');
      open my $fh, '<', $dbConfig_file || warn "Failed to open DB configuration file $dbConfig_file: $!";
      $dbConfig={};
      $dbConfigTemp={};
      print STDERR "Reading DB configuration.\n";
      while (<$fh>) {
	chomp;                  # no newline
	s/#.*//;                # no comments
	s/^\s+//;               # no leading white
	s/\s+$//;               # no trailing white
	next unless length;     # anything left?
	my ($var, $value) = split(/\s*=\s*/, $_, 2);
	$dbConfig->{$var} = $value;
	$dbConfigTemp->{$var} = $value;
      }
      close($fh);
    }
    return $dbConfig;
  }

  sub getDbConfigTemp {
    return $dbConfigTemp;
  }

  sub setDbConfig {
    $dbConfig = $_[0];
  }

  sub setDbConfigTemp {
    $dbConfigTemp =  $_[0];
  }

}

{
  # Build a trie-like data structure for the tagset
  # and return the top-level trie

  # If an argument is given, the trie is searched for the
  # corresponding partial tag and the corresponding
  # sub-trie is returned.
  # This is loaded when TrEd starts

  my $tagset_trie;
  sub tagset_trie {
    my ($tag_part)=@_;

    unless ($tagset_trie) {
      my @tagset = @{get_tagset()};
      $tagset_trie={};
      for my $tag (@tagset) {
	my $trie=$tagset_trie;
	while ($tag=~s/^([[:upper:]0-9][[:lower:]0-9])//) {
	  if (length $tag) {
	    $trie=$trie->{$1}||={'..'=>$trie};
	  } else {
	    $trie->{$1}=1;
	    last;
	  }
	}
	if (length($tag)) {
	  $trie->{$tag}=1;
	}
      }
    }
    my $trie = $tagset_trie;
    while (ref($trie) and defined($tag_part) and $tag_part=~s/^([[:upper:]0-9][[:lower:]0-9])//) {
      $trie=$trie->{$1};
    }
    if (ref($trie) and defined $tag_part and length $tag_part) {
      return $trie->{$tag_part};
    }
    return $trie;
  }
}

# hook to offer tagset when editing via conventional GUI
sub attr_choices_hook {
  my ($attr_path,$node,undef,$editor)=@_; 
  if ($attr_path eq 'tag') {
    return get_tagset();
  }
}

# hook to add additional styles to nodes matching certain conditions
sub node_style_hook {
  my ($node,$styles)=@_;
  if  ($unknownTokens{$node->{sentord}}) {
    print STDERR  $node->{form} . " -  " . $node->{tag}. "\n";
    AddStyle($styles, "NodeLabel", -dodrawbox=>"yes");
  }
}


##########################################
# hooks to offer validation against tagset

sub attr_validate_hook {
  my ($attr_value,$attr_path,$node,undef,$editor)=@_;
  if ($attr_path eq 'tag') {
    my @tagset = @{get_tagset()};
    if (grep {$_ eq $attr_value} @tagset) {
      return 1;
    }
    return 0;
  }
}

sub after_edit_attr_hook {
  my ($node,$attr_path,$data_changed)=@_;
  if (($attr_path eq 'tag') and ($data_changed) and
      !attr_validate_hook($node->{$attr_path},$attr_path,$node)) {
    Undo();
    SaveUndo("NOOP"); # Disallow the user revert our Undo with a Redo
    EditAttribute($node,$attr_path); # EditAttribute call to let the user re-edit the data
  }
}

sub after_edit_node_hook {
  my ($node,$data_changed)=@_;
  if (($data_changed) and
      !attr_validate_hook($node->{tag},'tag',$node)) {
    Undo();
    SaveUndo("NOOP"); # Disallow the user revert our Undo with a Redo
    EditAttribute($node); # EditAttribute call to let the user re-edit the data
  }
}
##########################################

# return all leaf nodes descending from the current trie
sub trie_leaf_nodes {
  my ($trie)=@_;
  return ((sort grep { !ref($trie->{$_}) } keys %$trie),
	  (map {
	    my $key = $_;
	    map $key.$_, trie_leaf_nodes($trie->{$key})
	  } sort grep { $_ ne '..' and ref($trie->{$_}) } keys %$trie)
	 );
}

#bind edit_tag to key t menu Edit tag
#bind check_tokens to key u menu Check which tokens are unknown to the lexicon
{
  my ($d,$l1,$l2,$scrollLex,$tag_part,$search_path,$trie,$return);
  my @lextags;
  my $dbConnection;
  my $dbConfig = getDbConfig();
  my $table = $dbConfig->{TABLE};
  my @tagset = @{get_tagset()};

  sub init_tag_editor {

    # clear
    $tag_part='';
    $search_path='';
    $return='';
    $trie=undef;
    if ($d) {
      # dialog already exists, just clear
      ToplevelFrame->Unbusy(); # TrEd shows busy cursor on existing widgets when running a macro
      $l1->delete(0,'end');
      $l2->delete(0,'end');
      $scrollLex->delete(0,'end');
      return;
    }

    # create widgets


    $d=ToplevelFrame->DialogBox(-title=> "Edit Tag",
				-buttons=> ["OK", "Cancel"]);

    my $f = $d->Frame()->pack(-side=>'top',-expand=>0,-fill=>'x');
    my $b = $f->Button(-text=>'<',-image=>main::icon($grp->{framegroup},'1leftarrow'))->pack(-side=>'left');
    my $e = $f->Label(-textvariable=>\$tag_part)->pack(-side=>'left',-expand=>0,-fill=>'x');
    my $e2 = $f->Label(-textvariable=>\$search_path,-foreground=>'blue')->pack(-side=>'left',-expand=>0,-fill=>'x');
    my $f2 = $d->Frame()->pack(-side=>'top',-expand=>1,-fill=>'both');
    my $f3 = $d->Frame()->pack(-side=>'top',-expand=>2,-fill=>'both');
    my $f4 = $d->Frame()->pack(-side=>'top',-expand=>2,-fill=>'both');
    my $e4;


    if (getDbConnection()) {
       $e4 = $f4->Label(-text=>"Tag suggestions from lexicon DB",-foreground=>'blue')->pack(-side=>'top',-expand=>0,-fill=>'x');
     } else {
       $e4 = $f4->Label(-text=>"Lexicon DB unavaible. To use, please configure and restart TrEd.",-foreground=>'red')->pack(-side=>'top',-expand=>0,-fill=>'x');
     }
    $l1=$f2->Scrolled('Listbox',-scrollbars=>'oe',-selectmode=>'browse')->pack(-side=>'left',-expand=>1,-fill=>'both');
    $l2=$f2->Scrolled('Listbox',-scrollbars=>'oe',-selectmode=>'browse')->pack(-side=>'left',-expand=>1,-fill=>'both');
    $scrollLex=$f4->Scrolled('Listbox',-scrollbars=>'oe',-selectmode=>'browse')->pack(-side=>'left',-expand=>1,-fill=>'both');

    $_->Subwidget('yscrollbar')->configure(-takefocus=>0) for $l1, $l2;

    # keyboard bindings
    $l1->bind('<BackSpace>' => sub {
		if (length($search_path)) {
		  $search_path=substr($search_path,0,length($search_path)-1);
		} elsif ($l1->get(0) eq '..') {
		  $l1->activate(0);
		  $l1->see(0);
		  select_l1_item();
		}
		Tk->break;
	      });
    $l2->bind('<BackSpace>' => sub {
		$l1->focus;
		if (length($search_path)) {
		  $search_path=substr($search_path,0,length($search_path)-1);
		} elsif ($l1->get(0) eq '..') {
		  $l1->activate(0);
		  $l1->see(0);
		  select_l1_item();
		}
		Tk->break;
	      });
    $b->configure(-command => sub {
		    if ($d->focusCurrent==$l2) {
		      $l2->bind('<BackSpace>')->Call;
		    } else {
		      $l1->bind('<BackSpace>')->Call;
		    }
		  });

    for my $w (map $_->Subwidget('scrolled'), $l1,$l2,  $scrollLex) {
      $w->bindtags([$w,ref($w),$w->toplevel,'all']);

      # select by typing
      $w->bind($w,'<KeyPress>',
	       sub {
		 my ($w)=@_;
		 my $key = $w->XEvent->K;
		 my @values = $w->get(0,'end');
		 my @found = grep { index(lc($values[$_]),lc($search_path.$key))==0 } 0..$#values;
		 if (@found) {
		   my $index= $found[0];
		   $search_path=substr($values[$index],0,length($search_path.$key));
		   $w->selectionClear(0,'end');
		   $w->selectionSet($index);
		   $w->activate($index);
		   $w->see($index);
		   if (@found==1) {
		     # call whatever si bound to Return
		     Tk::catch { $w->bind('<Return>')->Call() };
		   }
		 }
		 Tk->break if ($key=~/^[[:alnum:]]$/);
	       });

      $w->bind('<FocusIn>' => sub {
		 my ($w)=@_;
		 $w->selectionClear(0,'end');
		 $w->selectionSet('active');
	       });
      $w->bind('<1>' => sub {
		 my ($w)=@_;
		 $w->focus;
		 $w->activate('@'.$w->XEvent->x.",".$w->XEvent->y);
		 $w->see('active');
		 $w->selectionClear(0,'end');
		 $w->selectionSet('active');
		 Tk->break;
	       });
    }

    for my $event (qw(<Double-1> <space> <Return>)) {
      $l1->Subwidget('scrolled')->bind($event => \&select_l1_item);
      $l2->Subwidget('scrolled')->bind($event => sub {
					 my $active = $l2->get('active');
					 if (defined($active) and length($active)) {
					   $return=$tag_part.$active;
					   $d->{selected_button}='OK';
					 }
					 Tk->break;
				       });
      $scrollLex->Subwidget('scrolled')->bind($event => sub {
					 my $active = $scrollLex->get('active');
					 if (defined($active) and length($active)) {
					   $return=$active;
					   $d->{selected_button}='OK';
					 }
					 Tk->break;
				       });

    }
    $d->configure(-focus=>$l1);

    $d->BindReturn($d,1);
    $d->BindEscape();

    $d->Subwidget('B_OK')->configure(
      -command=> sub {
	my $append; # item in the focused list
	if ($l2->Subwidget('scrolled')==$d->focusCurrent) {
	  $append=$l2->get('active');
	} elsif ($l1->Subwidget('scrolled')==$d->focusCurrent) {
	  $append=$l1->get('active');
	} elsif ($scrollLex->Subwidget('scrolled')==$d->focusCurrent) {
	  $tag_part=$scrollLex->get('active');
	}

	if (($scrollLex->Subwidget('scrolled')==$d->focusCurrent) && (tagset_trie($tag_part)==1)) {
	  $return=$tag_part;
	  $d->{selected_button}='OK';
	} elsif (tagset_trie($tag_part.$append)==1) {
	  $return = $tag_part.$append;
	  $d->{selected_button}='OK';
	} elsif (tagset_trie($tag_part)==1) {
	  $return = $tag_part;
	  $d->{selected_button}='OK';
	}
	Tk->break;
      });
  }

  sub select_l1_item {
    $l1->focus;
    my $anchor=$l1->get('active');

    if ($anchor eq '..') {
      $tag_part=substr($tag_part,0,length($tag_part)-2);
    } elsif (defined $anchor and length $anchor) {
      $tag_part.=$anchor;
    } else {
      Tk->break;
      return;
    }
    $search_path='';
    $trie = $trie->{$anchor};
    if (!ref($trie)) {
      # a leaf (full tag completion)
      if (defined($anchor) and length($anchor)) {
	$return=$tag_part;
	$d->{selected_button}='OK';
      }
      Tk->break;
    }

    my @keys = keys(%$trie);
    if ($anchor ne '..') {
      # if there is just one partial completion, use it automatically
      while (@keys<=2
	     and grep(!ref($trie->{$_}),@keys)==0
	     and grep($_ ne '..',@keys)==1) {
	($anchor) = grep { $_ ne '..' } @keys;
	$tag_part.=$anchor;
	$trie = $trie->{$anchor};
	@keys = keys(%$trie);
      }
    }

    $l1->selectionClear(0,'end');
    $l1->delete(0,'end');
    $l1->insert('end', grep { $tag_part || !/^[[:upper:]][[:upper:]]/ } sort @keys );
    my $index = $l1->get(0) eq '..' ? 1 : 0;
    $l2->selectionClear(0,'end');
    $l2->delete(0,'end');
    $l2->insert('end', trie_leaf_nodes($trie) );
    if ($index>=$l1->size) {
      $l2->focus;
      $l2->selectionSet(0);
      $l2->activate(0);
      $l2->see(0);
    } else {
      $l1->selectionSet($index);
      $l1->activate($index);
      $l1->see(0);
    }
    Tk->break;
  }


  sub edit_tag {
    # initialize the Dialog GUI

    init_tag_editor();

    my $tag = $this->{tag};
    my $word = $this->{form};


    if (getDbConnection()) {
      eval {@lextags = @{do_select_stmt("SELECT Tag FROM $table WHERE Word = \'$word\';") }};
      if (!@lextags) {
	my $wordTemp = lc($word);
	eval {@lextags = @{do_select_stmt( "SELECT Tag FROM $table WHERE Word = \'$wordTemp\';") }};
      }
    }

    @lextags = sort(@lextags);

    foreach my $lextag (@lextags) {
      # Validate
      if (grep {$_ eq $lextag} @tagset) {
	$scrollLex->insert('end', $lextag );
      }
    }

    $trie=undef;
    if ($tag) {
      $tag_part=$tag;
      while (!ref($trie) and length($tag_part)>=2) {
	$tag_part = substr($tag_part,0,length($tag_part)-2);
	$trie = tagset_trie($tag_part);
      }
      if (!ref($trie)) {
	$tag_part='';
	$trie=undef;
      }
    }
    $trie||=tagset_trie();

    $l1->insert('end',
		grep { $tag_part || !/^[[:upper:]][[:upper:]]/ }
		  sort keys %$trie );
    $l1->activate(0);
    $l1->see(0);
    $l1->selectionSet(0);
    $l2->insert('end', trie_leaf_nodes($trie) );

    if ($trie->{substr($tag,length($tag_part))}==1) {
      # select the current tag in the left view
      my $index=Index([$l2->get(0,'end')],substr($tag,length($tag_part)));
      $l2->focus;
      $l2->activate($index);
      $l2->selectionSet($index);
      $l2->see($index);
    }


    if ($d->Show eq 'OK') {
      if (defined($return) and length($return)) {
	$this->{tag} = $return;
      }
    }
  }

  sub check_tokens {

    if (getDbConnection()) {
	my @nodes = ();
	my $node = $root;

	while ($node) {

	  $node = $node->following($root);
	  if ($node) {
	    if (!$node->attr("form")) {
	      next;
	    } elsif ($node->attr("tag") eq "-") {
	      next;
	    } else {
	      push(@nodes, $node);
	    }
	  }
	}

	@nodes = sort {$a->attr("sentord") <=> $b->attr("sentord") } @nodes;
	foreach (@nodes) {
	  my @tags = ();
	  my $word=$_->attr("form");
	  eval {@tags = @{do_select_stmt("SELECT Tag FROM $table WHERE Word = \'$word\';") }};
	  if (!@tags) {
	    my $wordTemp = lc($word);
	    eval {@tags = @{do_select_stmt( "SELECT Tag FROM $table WHERE Word = \'$wordTemp\';") }};
	  }
	  if (!@tags) {
	    $unknownTokens{$_->{sentord}} = 1;
	  }
	}
	Redraw() if GUI();
	%unknownTokens=();
      }

    # Prohibit redrawing again, with %unknownTokens empty
    $FileChanged=0;
    $Redraw="none";
  }

  sub do_select_stmt {
    my $stmt= shift;
    my $dbConnection = getDbConnection();
    my @lextags;

    my $sth = $dbConnection->prepare($stmt) or print STDERR $DBI::errstr;
    $sth->execute or print STDERR $DBI::errstr;
    my @results;
    while (@results = $sth->fetchrow_array) {
      push(@lextags, $results[0]);
    }
    $sth->finish;
    return \@lextags;
  }

}


#bind autoCheckLemmas to key l menu Automatically check lemmas for possible errors
{

  my ($d, $l1);
  my @nodes;
  my $lemmatizer;
  my $dbConnection;
  my %lemRecs;
  my %origLemmas;
  my %currentLemmas;

  sub init_lemma_check_infobox {
    # printStatus();

    if ($d) {
       # dialog already exists, just clear
       ToplevelFrame->Unbusy(); # TrEd shows busy cursor on existing widgets when running a macro
     }

    # create widgets
    $d=ToplevelFrame->DialogBox(-title=> "Automatically checking lemmas for possible errors",
				#-buttons=> ["Previous","Close","Undo", "Save", "Next"]);
				-default_button => 'Save',
				-buttons=> [ "Prev", "Close","Undo", "Save", "Next"]);


    if ($grp->{treeNo} <= 0) {
      $d->Subwidget('B_Prev')-> configure('-state' => 'disabled');
    } elsif ($grp->{treeNo} >= $grp->{FSFile}->lastTreeNo) {
      $d->Subwidget('B_Next')-> configure('-state' => 'disabled');
    }


    my $f1 = $d->Labelframe( -padx=>2,    -pady=>2)->pack(-side=>'top',-expand=>1,-fill=>'both');
    $dbConnection = getDbConnection();



    if (!$dbConnection ) {
      $l1=$f1->Scrolled('Listbox',-scrollbars=>'oe',-width=>80,-height=>10,-setgrid=>1, -scrollbars=>'se', -selectmode=>'browse')
	->pack(-side=>'top',-expand=>1,-fill=>'both');

      $l1->delete('0.0','end');
      $l1->insert( 'end', "Lexicon DB unavaible. To use, please configure and restart TrEd.");
      $l1->focus;
    } else {
      my $table;
      $table = $f1->Table(-rows => $#nodes + 4,
			  -fixedrows => 1,

			  -scrollbars => 'se',
			  -takefocus => 1,
			  -relief => 'raised')->pack(-side => 'bottom' ,-expand=>1,-fill=>'both');


      #$table->minsize($table->winfo('width'), $table->winfo('height'));


      my $idLabel = $table->Label(-text => "Id", -width => 6, -relief =>'raised');
      $table->put(0, 1, $idLabel);

      my $formLabel = $table->Label(-text => "Form", -width => 24, -relief =>'raised');
      $table->put(0, 2, $formLabel);

      my $lemmaLabel = $table->Label(-text => "Lemma", -width => 24, -relief =>'raised');
      $table->put(0, 3, $lemmaLabel);


      my $row = 1;
      foreach my $node (@nodes) {

	my $nodeIdLabel = $table->Label(-text => ($row) ,
					-padx => 2,
					-anchor => 'w',
					-background => 'white',
					-relief => "groove");
	$table->put($row, 1, $nodeIdLabel);



	my $formLabel = $table->Label(	    -text => ($node->attr("form")) ,
					    -padx => 2,
					    -anchor => 'w',
					    -background => 'white',
					    -relief => "groove");
	$table->put($row, 2, $formLabel);

	my $lemmaF = $table->Frame->pack( qw/-side left/ );



	my $lemRec = $lemRecs{$node->attr("sentord")};

	if ($lemRec->{"OTHERLEMMAS"} == 0) {

	  my $lemEnt =  $lemmaF->Entry(-textvariable=>\$currentLemmas{$node->attr("sentord")})->pack(-side=>'left',-expand=>0,-fill=>'x');
	  if ($lemRec->{"LEMMAINDB"} != 1) {
	    $lemEnt->configure( -background => 'PapayaWhip', -foreground => 'black');
	    # print STDERR "NOTFOUND " . $node->attr("form") . "\n";
	  }


	} else {

	  my $lemBE = $lemmaF->BrowseEntry( -variable => \$currentLemmas{$node->attr("sentord")})->pack( qw/-side left/,-expand=>0,-fill=>'x' );

	  if ($lemRec->{"FOUND"} == 1) {
	    # print STDERR "FOUND " . $node->attr("form") . "\n";
	  } else {
	    $lemBE->configure( -background => 'PapayaWhip', -foreground => 'black');
	    # print STDERR "NOTFOUND " . $node->attr("form") . "\n";
	  }

	  $lemBE->insert("end",  $node->attr("lemma"));

	  # print Dumper $lemRec;

	  foreach ((keys %{$lemRec->{"SAMETAGLEMMAS"}}) , (keys %{$lemRec->{"OTHERTAGLEMMAS"}}) ) {
	    # print STDERR "LEMMA " . $_ . "\n";
	    $lemBE->insert("end",  $_);
	  }

	  $lemBE->configure( -background => 'white', -foreground => 'black');

	}

	$table->put($row, 3, $lemmaF);

	$row++;
      }
      $table->pack(-expand=>'yes');

    }


    # FIXME $d->{selected_button}='Save';

    $d->configure();
    $d->BindReturn($d,1);
    $d->BindEscape();


    $d->Subwidget('B_Prev')->configure(
				       -command=> sub {
					 &myPrevTree;
					 Redraw() if GUI();
					 $d->destroy;
					 $root = $grp->{root};
					 &clearTreeInfo;
					 &autoCheckLemmas;
				       }
				      );



    $d->Subwidget('B_Close')->configure(
					-command=> sub {
					  $d->destroy;
					});

    $d->Subwidget('B_Undo')->configure(
				       -command=>  sub {
					 foreach my $node (@nodes) {
					   # print STDERR  $node->attr("lemma") . " " . $origLemmas{$node->attr("sentord")} . "\n";
					   $node->set_attr("lemma", $origLemmas{$node->attr("sentord")});
					   $currentLemmas{$node->attr("sentord")} = $node->attr("lemma");
					 }
					 Redraw() if GUI();
					});


    $d->Subwidget('B_Save')->configure(
				       -command=> sub {
					 foreach my $node (@nodes) {
					   $origLemmas{$node->attr("sentord")} = $node->attr("lemma");
					   $node->set_attr("lemma", $currentLemmas{$node->attr("sentord")});
					 }
					 Redraw() if GUI();
				       });



    $d->Subwidget('B_Next')->configure(
				       -command=> sub {
					 &myNextTree;
					 Redraw() if GUI();
					 $d->destroy;
					 $root = $grp->{root};
					 &clearTreeInfo;
					 &autoCheckLemmas;
				       }
				      );


  }


  sub clearTreeInfo {
    @nodes = ();
    %lemRecs = ();
    %origLemmas = ();
    %currentLemmas = ();

  }

  sub check_lemmas {

    $lemmatizer = getLemmatizer();

    # No lemmas found for this token: Red
    # No lemmas found for this token + tag: LightRed
    # Multiple lemmas found for this token: Brown
    # Multiple lemmas found for this token + tag: Blue
    # One lemma found for this token: Black

    %origLemmas = ();
    %currentLemmas = ();

    @nodes = ();

    my $node = $root;

    while ($node) {
      my $lemRec = undef;

      $node = $node->following($root);
      if ($node) {
	if (!$node->attr("form")) {
	  next;
	} elsif ($node->attr("tag") eq "-") {
	  next;
	} else {

	  $origLemmas{$node->attr("sentord")}= $node->attr("lemma");
	  $currentLemmas{$node->attr("sentord")}= $node->attr("lemma");

	  push(@nodes, $node);

	  if (defined ($node->attr("lemma"))) {
	    $lemRec->{"ORIG_LEMMA"} = $node->attr("lemma");
	  } else {
	    $lemRec->{"ORIG_LEMMA"} = $node->attr("form");
	  }

	  $lemRec->{"FOUND"} = 0 ;
	  $lemRec->{"LEMMAINDB"} = 0 ;
	  $lemRec->{"OTHERLEMMAS"} = 0 ;

	  $lemmatizer->getLemmas($node, $lemRec);

	  $lemRecs{$node->attr("sentord")}= $lemRec;

	  # print Dumper  $lemRec;

	  #$node->set_attr("lemma");
	  #print STDERR $lemmas;
	}
      }

    }
    @nodes = sort {$a->attr("sentord") <=> $b->attr("sentord") } @nodes;

  }

  sub autoCheckLemmas {
    check_lemmas();
    init_lemma_check_infobox();
    $d->Show;
  }

}

#bind configureLexDbSetts to key c menu Configure lexicon DB settings
{
  my ($d,$l1,$return);
  my $dbConfig = getDbConfig();
  my $dbConfigTemp = getDbConfigTemp();

  my %labEntries;

  sub init_config_editor {

    # printStatus();

    if ($d) {
       # dialog already exists, just clear
       ToplevelFrame->Unbusy(); # TrEd shows busy cursor on existing widgets when running a macro
     }

    # create widgets
    $d=ToplevelFrame->DialogBox(-title=> "Lexicon Database Settings",
				-buttons=> ["Save",  "Test", "Cancel"]);

    my $f1 = $d->Labelframe(-text=>"DB Connection Status", -padx=>2, -pady=>2)->pack(-side=>'top',-expand=>1,-fill=>'both');

    $l1=$f1->Scrolled('Listbox',-scrollbars=>'oe',-selectmode=>'browse')->pack(-side=>'left',-expand=>1,-fill=>'both');

    $_->Subwidget('yscrollbar')->configure(-takefocus=>0) for $l1;

    my $f2 = $d->Frame()->pack(-side=>'top',-expand=>1,-fill=>'both');
    my $dbConnection = getDbConnection();
    $l1->delete('0.0','end');
    if ($dbConnection ) {
      $l1->insert( 'end', "Connection with lexicon DB established." );
    } else {
      $l1->insert( 'end', getError() );
    }

    foreach my $item (
		      ['User', \$dbConfig->{USER}],
		      ['Password', \$dbConfig->{PWD}],
		      ['Host',  \$dbConfig->{HOST}],
		      ['Database Server', \$dbConfig->{DBSERVER} ],
		      ['Database', \$dbConfig->{DATABASE}],
		      ['Table', \$dbConfig->{TABLE}],
		      ['Port', \$dbConfig->{PORT} ],
		     ) {
      my $l = $item->[0] . ':';
      my $le = $f2->LabEntry(
			     -label        => ' ' x (13 - length $l) . $l,
			     -labelPack    => [qw/-side left -anchor w/],
			     #-labelFont    => '9x15bold',
			     # -relief       => 'flat',
			     #-state        => 'disabled',
			     -textvariable => $item->[1],
			     -width        => 35,
			    );
      if ($le->cget('label') =~ "Password") {
	$le->configure(-show => 0);
      }
      $le->pack(qw/-fill x -expand 1/);

      # Remember %labEntries;
      $labEntries{$item->[0]} = $le;
    }


    $d->configure(-focus=>$l1);

    $d->BindReturn($d,1);
    $d->BindEscape();

    $d->Subwidget('B_Save')->configure(
      -command=> sub {
	writeDbConfig();
	setDbConfig($dbConfig);
	setDbConfigTemp($dbConfig);
	$d->destroy;
      });

    # $l1->tag(qw/configure error -foreground red/);

    $d->Subwidget('B_Test')->configure(
      -command=> sub {
	resetDbConnection();
	# printStatus();
	setDbConfig($dbConfig);
	# printStatus();
	my $dbConnection = getDbConnection();
	# printStatus();
	$l1->delete('0.0','end');
	if ($dbConnection ) {
	  $l1->insert( 'end', "Connection with lexicon DB established." );
	} else {
	  $l1->insert( 'end', getError() );
	}

      });

    $d->Subwidget('B_Cancel')->configure(
      -command=> sub {

	resetDbConnection();
	setDbConfig($dbConfigTemp);
	my $dbConfig = getDbConfig();
	my $dbConnection = getDbConnection();
	# printStatus();
	$l1->delete('0.0','end');
	if ($dbConnection ) {
	  $l1->insert( 'end', "Connection with lexicon DB established." );
	} else {
	  $l1->insert( 'end', getError() );
	}

	foreach my $item (
			  ['User', \$dbConfig->{USER}],
			  ['Password', \$dbConfig->{PWD}],
			  ['Host',  \$dbConfig->{HOST}],
			  ['Database Server', \$dbConfig->{DBSERVER} ],
			  ['Database', \$dbConfig->{DATABASE}],
			  ['Table', \$dbConfig->{TABLE}],
			  ['Port', \$dbConfig->{PORT} ],
			 ) {
	  # Fixme: how can I reset on Cancel?
	  # ($labEntries{$item->[0]})->configure(-textvariable => $item->[1]);
	  # print STDERR "LABEL ".$labEntries{$item->[0]}->cget('textvariable') . "\n";
	  # $d->update;
	}
	$d->destroy;
      });
  }

  sub configureLexDbSetts {
    # initialize the Dialog GUI
    init_config_editor();
    $d->Show;
  }

  # write the DB configuration settings in the DB config file
  # (loading from source file just once)
  sub writeDbConfig {
    my $dbConfig_file = FindInResources('gdt/dbConfig.txt');
    open my $fh, '>', $dbConfig_file || warn "Cannot write to DB configuration file $dbConfig_file: $!";
    while (my ($var, $value)   = each(%{$dbConfig})){
      print STDERR  ($var ." = " . $value . "\n");
      print $fh  ($var ." = " . $value . "\n");
    }
    close($fh);
  }

}


{
  # return the DB connection
  # This is loaded when TrEd starts
  my $dbConnection;
  my $error;

  sub getDbConnection {
    unless ($dbConnection) {
      my $dbConfig = getDbConfig();
      my $data_source;warn $@ if $@;

      if ($dbConfig->{DBSERVER} !~ /^postgre/i) {
	$error = "ERROR: Connection unavailable: Only postgres db server supported";
      } else {

	my $database= $dbConfig->{DATABASE};
	my $host= $dbConfig->{HOST};
	my $port= $dbConfig->{PORT};
	my $user = $dbConfig->{USER};
	my $password = $dbConfig->{PWD};

	$dbConnection = DBI->connect("dbi:Pg:dbname=$database;host=$host;port=$port;", "$user", "$password");

	if (defined $dbConnection) {
	  $error = undef;
	} else {
	  $error = $DBI::errstr;
	  print STDERR $error;
	}

	if (defined $dbConnection) {
	  my $tExists = undef;
	  $tExists = checkTable();
	  # my $tExists = 0;
	  # my $tempTable = "public.".$dbConfig->{TABLE};
	  # # print STDERR "$tempTable\n";
	  # foreach my $table ( $dbConnection->tables()) {
	  #   print STDERR "$table\n";
	  #   if ($table =~ /^$tempTable$/i) {
	  #     $tExists = 1;
	  #   }
	  # }
	  if ($tExists) {
	    $error = undef;
	  } else {
	    $error = "ERROR: Cannot access table " .$dbConfig->{TABLE};
	    $dbConnection = undef;
	  }
	}

      }
    }
    return $dbConnection;
  }

  sub checkTable {
    my @results;
    if (getDbConnection()) {
      my $dbConfig = getDbConfig();
      eval {@results = @{do_select_stmt("SELECT Tag FROM " . $dbConfig->{TABLE} . " WHERE Word = \'dummy\';") }};
      unless ($@) {
	return 1;
      }
    }
    return undef;
  }

  sub resetDbConnection {
    if ($dbConnection) {
      $dbConnection->disconnect;
    }
    $dbConnection = undef;
  }

  sub getError {
    return $error;
  }


}

{
  # return the lemmatizer
  # This is loaded when TrEd starts
  my $lemmatizer;
  my $error;

  sub getLemmatizer {
    unless ($lemmatizer) {
      my $dbConnection = getDbConnection();
      $lemmatizer = Lemmatizer->new();
      $lemmatizer->setDbConnection($dbConnection);
      $lemmatizer->setDebugStatus(1);
    }

    return $lemmatizer;
  }
}



{
  sub printStatus {
    print STDERR "TMP DB Config is...\n";
    print Dumper getDbConfigTemp();
    print STDERR "DB Config is...\n";
    print Dumper getDbConfig();
    print STDERR "Connection is...\n";
    print Dumper getDbConnection();
  }
}

{
  sub myNextTree {
    return 0 if ($grp->{treeNo} >= $grp->{FSFile}->lastTreeNo);
    gotoTree($grp,$grp->{treeNo}+1);
    return 1;
  }


  sub myPrevTree {
    return 0 if ($grp->{treeNo} <= 0);
    gotoTree($grp,$grp->{treeNo}-1);
    return 1;
  }


}

1;
