#============================================================
#
# Lemmatizer.pm
#
# DESCRIPTION
#   Retrieve lemma from postgres ILSP lexicon
#
# AUTHOR
#   Prokopis Prokopidis  <prokopis@ilsp.gr>
#
# COPYRIGHT
#   Copyright (C) 2010 ILSP.  All Rights Reserved.
#
# #====================================================================
package Lemmatizer;


use strict;
use DBI;
use LexisConflate;	
use Encode;
use utf8;


my $greek_letters = "αβγδεζηθικλμνξοπρσςτυφχψωάέήίόύώϊϋΐΰΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩΆΈΉΊΌΎΏΪΫ";

# Capital - Small Letters
my $capital_greek_letters = "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩΆΈΉΊΌΎΏΪΫ";
my $small_greek_letters = "αβγδεζηθικλμνξοπρσςτυφχψωάέήίόύώϊϋΐΰ";

# Consonants
my $greek_consonants = "βγδζθκλμνξοπρσςτφχψωΒΓΔΖΘΚΛΜΝΞΠΡΣΤΦΧΨ";
my $capital_greek_consonants = "ΒΓΔΖΘΚΛΜΝΞΠΡΣΤΦΧΨ";
my $small_greek_consonants = "βγδζθκλμνξοπρσςτφχψ";

# Vowels
my $greek_vowels = "αεηιουωάέήίόύώϊϋΐΰΑΕΗΙΟΥΩΆΈΉΊΌΎΏΪΫ";
my $capital_greek_vowels = "ΑΕΗΙΟΥΩΆΈΉΊΌΎΏΪΫ";
my $small_greek_vowels = "αεηιουωάέήίόύώϊϋΐΰ";

my $greek_vowels_with_accents_and_diacritics = "άέήίόύώϊϋΐΰΆΈΉΊΌΎΏΪΫ";
my $small_greek_vowels_with_accents_and_diacritics = "άέήίόύώϊϋΐΰ";
my $capital_greek_vowels_with_accents_and_diacritics = "ΆΈΉΊΌΎΏΪΫ";

my $greek_vowels_with_accents = "άέήίόύώΆΈΉΊΌΎΏ";
my $small_greek_vowels_with_accents = "άέήίόύώ";
my $capital_greek_vowels_with_accents = "ΆΈΉΊΌΎΏ";

my $small_greek_vowels_without_accents = "αεηιουωϊϋ";
my $capital_greek_vowels_without_accents = "ΑΕΗΙΟΥΩΪΫ";
my $greek_vowels_without_accents = $small_greek_vowels_without_accents . $capital_greek_vowels_without_accents ;
my $capital_greek_letters_without_accents = "ΑΕΗΙΟΥΩΪΫΒΓΔΖΘΚΛΜΝΞΠΡΣΤΦΧΨ";
my $small_greek_letters_without_accents = "αεηιουωϊϋβγδζθκλμνξοπρσςτφχψ";
my $greek_letters_without_accents = $small_greek_letters_without_accents . $capital_greek_letters_without_accents ;

my $greek_vowels_with_diacritics = "ϊϋΐΰΪΫ";
my $small_greek_vowels_with_diacritics = "ϊϋΐΰ";
my $capital_greek_vowels_with_diacritics = "ΪΫ";

# Some regexes for the $unaccented_first_capital_problem as in, for example, ENET files
my $unaccented_first_capital_regex = "[$capital_greek_vowels_without_accents]+[$small_greek_consonants]+[$small_greek_vowels_without_accents]+[$small_greek_letters_without_accents]*";


my %fields = (
    dbConnection=>undef,
    debug=>undef,
    );


my %DEFAULTS = (
    '1'  => "No",
    '2'   => "Aj",
    '3'    => "Pn",
    '4'  => "At",
    '5'   => "Ad",
    '6'    => "Vb",
    '7'  => "Cj",
    '8'   => "As",
    '9'    => "Ij",
    '10'  => "",
    '11'   => "Rg",
    '12'    => "Pt",
    '13'  => "Nm",
    '14'   => "",
    '15'    => "",
    '16'  => "Nm",
    '19'  => "Rg",
);



my %NUMBER_DEFAULTS = (
    'Sg'  => "1",
    'Pl'   => "2",
);

my %GENDER_DEFAULTS = (
    'Ma'  => "1",
    'Fe'   => "2",
    'Ne'    => "3",
);

my %CASE_DEFAULTS = (
    'Nm'  => "1",
    'Ge'   => "2",
    'Ac'    => "3",
    'Vo'    => "4",
    'Da'    => "5",
);

my %PERSON_DEFAULTS = (
    '01'  => "1",
    '02'   => "2",
    '03'   => "3",
);

my %VOICE_DEFAULTS = (
    'Av'  => "1",
    'Pv'   => "2",
);

my %ASPECT_DEFAULTS = (
    'Av'  => "1",
    'Pv'   => "2",
);

my %PRONOUN_DEFAULTS = (
    'Pe'  => "0",
    'Po'  => "1",
    'Dm'  => "2",
    'Id'  => "3",
    'Ir'  => "4",
    'Re'  => "5",
    'Ri'  => "7",
    'PeWe'  => "9",

);


sub new {
    my ($class) = @_;
    my $self = {%fields};
    bless ($self, $class);  
    return $self;
}


sub setDbConnection { 
    my $self = shift;
    $$self{"dbConnection"}=shift;
    return $self;
}


sub getDbConnection { 
    my $self = shift;
    return $$self{"dbConnection"};
}

sub setDebugStatus { 
    my $self = shift;
    $$self{"debug"}=shift;
    return $self;
}

sub getDebugStatus { 
    my $self = shift;
    return $$self{"debug"};
}


sub getLemmas {

    my $self = shift;
    my $node = shift;
    my $lemRec = shift;
    my $debug= $self->getDebugStatus();

    my $word = $node->attr("form");
    my $tag =  $node->attr("tag");

    $self->getLemmasForVariants($lemRec, $word, $tag);

    if ( $lemRec->{"FOUND"} == 0 ) {

	foreach my $variant (@{$self->generateVariants($word)}) {
	    $self->getLemmasForVariants($lemRec, $variant, $tag);
	}
    }

    return $self;
}    

sub getLemmasForVariants {

    my $self = shift;
    my $lemRec = shift;
    my $word = shift;
    my $tag = shift;

    my $debug= $self->getDebugStatus();

    my $sql;

    my $base = substr($tag,0,2);


    if ($word =~ /[\'\"\;\:,.\/\?\<\>\~\`\!\@\#\$\%\^\^\%\^\&\*\(\)\_\-\+\=\|\\\{\[\}\]]/ ) {
	$lemRec->{"FOUND"} = 1;
	$lemRec->{"LEMMAINDB"} = 1;

    } elsif ($word =~ /^\d+/ ) {
	$lemRec->{"FOUND"} = 1;
	$lemRec->{"LEMMAINDB"} = 1;
    } else {
	
	my $normalized_word = &LexisConflate::remove_englitikos_tonos($word);


	# Check if the word exists in the DB with its current orthography
	# Incorporate subPOS information
	$sql = $self->build_sql($normalized_word, $tag);
	# print STDERR  "$sql " .$lemRec->{"FOUND"} . "\n"  ;    
	eval {$self->do_select_stmt($sql, $base,  $lemRec)};
	# print STDERR  "$sql " .$lemRec->{"FOUND"} . "\n"  ;    
	
	# Without any subpos information
	if ($lemRec->{"FOUND"} == 1) {
	    $sql =  $self->build_sql($normalized_word);
	    # print STDERR  "$sql\n"  ;    
	    eval {$self->do_select_stmt($sql, $base, $lemRec)};
	    $self->show_err() if $@;
	} 
	
	
	# If not, decapitalize and try with an accented version
	# Incorporate subPOS information
	if ($lemRec->{"FOUND"} == 0) {
	    my $wordsmall = $normalized_word;
	    if (( $normalized_word =~ /^[ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ]/ )  && (length($normalized_word) gt 1) ) {
		# print STDERR "NORMWORD = $normalized_word \n" if ($debug == 1);
		$wordsmall =~ s/Α$/ά/;
		$wordsmall =~ s/Ε$/έ/;
		$wordsmall =~ s/Ι$/ί/;
		$wordsmall =~ s/Ο$/ό/;
		$wordsmall =~ s/Υ$/ύ/;
		$wordsmall =~ s/Η$/ή/;
		$wordsmall =~ s/Ω$/ώ/;
		$wordsmall =~ s/Σ$/ς/;
		# print STDERR "WORDSMALL = $wordsmall \n" if $debug ==1;
	    }
	    
	    $wordsmall =~ tr/[ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩΆΌΈΎΊΉΏ]/[αβγδεζηθικλμνξοπρστυφχψωάόέύίήώ]/;
	    # print STDERR  "WORDSMALL = $wordsmall \n" if $debug ==1;
	    
	    $sql = $self->build_sql($wordsmall, $tag);
	    
	    #$text2 = "SELECT Lemma,Lemgram FROM Entries WHERE Word = \'$wordsmall\';"  ;
	    #$text2 = "SELECT LEMMA,LEMGRAM FROM FULL WHERE WORD=\'" . "$wordsmall" . "\';"  ;
	    eval {$self->do_select_stmt($sql, $base, $lemRec)};
	    $self->show_err() if $@;
	    
	    # Without any subpos information
	    if ($lemRec->{"FOUND"} == 0) {
		$sql =  $self->build_sql($wordsmall);
		eval {$self->do_select_stmt($sql, $base, $lemRec)};
		$self->show_err() if $@;
	    } 	    	
	}
    }
}


sub build_sql {
   my $self= shift;
   my $word = shift;
   my $tag = shift;


   my $debug = $self->getDebugStatus();

   my $sql = "SELECT Lemma,Lemgram FROM Entries WHERE Word = \'$word\' "  ;  

   return $sql . " ;" unless defined($tag);

   my $pos = substr($tag,0,2);

   if (($pos eq "No") || ($pos eq "Aj") || ($pos eq "At")) {
       # The Gender constraint seems to deteriorate performance of the lemmatizer
       # because of wrong gender assignment by the tagger.
       # my $case = $CASE_DEFAULTS{substr($tag,8,2)};
       my $gender = $GENDER_DEFAULTS{substr($tag,4,2)};
       my $number = $NUMBER_DEFAULTS{substr($tag,6,2)};
       # The Case constraint seems to deteriorate performance of the lemmatizer
       # because of wrong case assignment by the tagger.
       # my $case = $CASE_DEFAULTS{substr($tag,8,2)};
       $sql .= "AND Gender = \'$gender\' "  if (defined($gender));
       $sql .= "AND Number = \'$number\' "  if (defined($number));
       # $sql .= "AND _Case = \'$case\' "  if (defined($case));

   } elsif ($pos eq "Pn") {
       my $tempsubtype = substr($tag,2,2);
       $tempsubtype .= "We" if (substr($tag,12,2) eq "We");
       my $subtype = $PRONOUN_DEFAULTS{$tempsubtype};
       # The gender constraint deteriorates performance of the lemmatizer
       # because of undefined gender (Gender = 4 ) for a lot of pronouns in the lexicon.
       #my $gender = $GENDER_DEFAULTS{substr($tag,4,2)};
       my $person = $PERSON_DEFAULTS{substr($tag,6,2)};
       my $number = $NUMBER_DEFAULTS{substr($tag,8,2)};
       $sql .= "AND Subtype = \'$subtype\' "  if (defined($subtype));
       #$sql .= "AND Gender = \'$gender\' "  if (defined($gender));
       $sql .= "AND Number = \'$number\' "  if (defined($number));
       
   }

   # print STDERR  "$word $tag \n$sql\;\n" if $debug == 1;
   
   return $sql . " ;";
}



sub do_select_stmt {
   my $self = shift;
   my $stmt = shift;
   my $tag = shift;
   my $lemRec = shift;

   my $dbh=$self->getDbConnection();
   my $debug = $self->getDebugStatus();

   my $lem;
   my $lextag;


   my $sth = $dbh->prepare($stmt) or die $DBI::errstr;
   $sth->execute or die $DBI::errstr;
   # print STDERR "STMT: $stmt\n" if $debug == 1;   

   if ($sth) {

       while (my @values = $sth->fetchrow_array) {
	   # I have found sth
	   $lemRec->{"FOUND"} = 1;
	   $lem = $values[0];
	   Encode::_utf8_on( $lem);

	   # print STDERR "$lem\n";

	   if ($lemRec->{"ORIG_LEMMA"} eq $lem) {
	       $lemRec->{"LEMMAINDB"} = 1;
	   } else {
	       $lemRec->{"OTHERLEMMAS"} = 1;
	       $lextag = $DEFAULTS{$values[1]};	
	       if ($lextag eq $tag) {
		   $lemRec->{"SAMETAGLEMMAS"}->{$lem} = 1 ;
	       } else {
		   $lemRec->{"OTHERTAGLEMMAS"}->{$lem} = 1 ;
	       }	   
	   }
       }

       $sth->finish;

   } else {
       $self->show_err() if $@;
   }

   print Dumper  $lemRec ;

}

sub _utf8_on_all_arrayref {
 
# For all the records specified
#  Switch on the UTF-8 flag for all values
# Return the original reference

    foreach (@{$_[0]}) {

        Encode::_utf8_on( $_ ) foreach @{$_};
    }
    $_[0];
} #_utf8_on_all_arrayref




# stuffs the contents of $dbh->Error
# into the output window after
# wrapping the error string
sub show_err {

    my $self=shift;
    my $dbh=$self->getDbConnection();
    print STDERR  $dbh->err, "error", "\n";
}


sub generateVariants {
    my $self=shift;
    my $word = shift;
    my @variants;

    $word = &LexisConflate::latin_greek_mixed($word);
    if ($word =~ /(.+)Σ$/) {
	$word = $1. "ς";
    }
    push(@variants, $word);
    my $ucfirst = ucfirst(lc($word));
    my $lc = lc($word);

    # Where are my vowels?
    my @chars = split(//, $word);
    my @unaccentedIdxs;
    for (my $i=0; $i<=$#chars; $i++) {
	if ($chars[$i] =~ /[$greek_vowels_without_accents]/) {
	    unshift( @unaccentedIdxs, $i);
	    # Use unshift instead of push so that vowels near the end of the word are checked first
	}
    }

    if ($word  =~ /^$unaccented_first_capital_regex$/) {
	my $vowel = substr($word, 0, 1);			
	$vowel =~ tr/ΑΕΗΙΟΥΩ/ΆΈΉΊΌΎΏ/;			
	push(@variants, ($vowel . substr($word, 1)));
	foreach my $idx (@unaccentedIdxs) {	    
	    next if ($idx == 0);
	    push(@variants, substr($word, 0, $idx) . accentuate_vowel(substr($word,$idx,1)) . substr($word, $idx+1));
	}
    } elsif ($word =~ /^[$greek_letters_without_accents]+(ς?)$/ ) {
	foreach my $idx (@unaccentedIdxs) {	    
	    # All vowels 
	    push(@variants,  substr($word, 0, $idx) . accentuate_vowel(substr($word,$idx,1)) . substr($word, $idx+1));
	    # All vowels ucfirst
	    push(@variants,  substr($ucfirst, 0, $idx) . accentuate_vowel(substr($ucfirst,$idx,1))  . substr($ucfirst, $idx+1));
	    # All vowels lc
	    push(@variants,  substr($lc, 0, $idx) .  accentuate_vowel(substr($lc,$idx,1))  . substr($lc, $idx+1));
	}
	push(@variants, $ucfirst);
	push(@variants, $lc);
    }
    return \@variants;
}

sub  accentuate_vowel {
    my $vowel = $_[0];
    $vowel =~ tr/αεηιουωϊϋΑΕΗΙΟΥΩ/άέήίόύώΐΰΆΈΉΊΌΎΏ/;
    return $vowel;
}

sub remove_englitikos_tonos {
    my $word = $_[0];
    return &LexisConflate::remove_englitikos_tonos($word);
}

sub latin_greek_mixed {
    my $word = $_[0];
    return &LexisConflate::latin_greek_mixed($word);
}


1;
